@@ -2,17 +2,22 @@
  * options.c :  routines for performing OPTIONS server requests
  *
  * ====================================================================
- * Copyright (c) 2000-2006, 2009 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
@@ -20,12 +25,19 @@
 
 #include "svn_pools.h"
 #include "svn_error.h"
+#include "svn_dirent_uri.h"
 #include "svn_private_config.h"
 #include "../libsvn_ra/ra_loader.h"
 
+#include "private/svn_fspath.h"
+
 #include "ra_neon.h"
 
 
+/* In a debug build, setting this environment variable to "yes" will force
+   the client to speak v1, even if the server is capable of speaking v2. */
+#define SVN_IGNORE_V2_ENV_VAR "SVN_I_LIKE_LATENCY_SO_IGNORE_HTTPV2"
+
 static const svn_ra_neon__xml_elm_t options_elements[] =
 {
   { "DAV:", "activity-collection-set", ELEM_activity_coll_set, 0 },
@@ -35,7 +47,7 @@ static const svn_ra_neon__xml_elm_t options_elements[] =
   { NULL }
 };
 
-typedef struct {
+typedef struct options_ctx_t {
   /*WARNING: WANT_CDATA should stay the first element in the baton:
     svn_ra_neon__xml_collect_cdata() assumes the baton starts with a stringbuf.
   */
@@ -102,7 +114,9 @@ end_element(void *baton, int state,
   options_ctx_t *oc = baton;
 
   if (state == ELEM_href)
-    oc->activity_coll = svn_string_create_from_buf(oc->cdata, oc->pool);
+    oc->activity_coll =
+      svn_string_create(svn_urlpath__canonicalize(oc->cdata->data, oc->pool),
+                        oc->pool);
 
   return SVN_NO_ERROR;
 }
@@ -110,23 +124,6 @@ end_element(void *baton, int state,
 
 /** Capabilities exchange. */
 
-/* Return TRUE iff STR exactly matches any of the elements of LIST. */
-static svn_boolean_t
-match_list(const char *str, const apr_array_header_t *list)
-{
-  int i;
-
-  for (i = 0; i < list->nelts; i++)
-    {
-      const char *this_str = APR_ARRAY_IDX(list, i, char *);
-
-      if (strcmp(this_str, str) == 0)
-        return TRUE;
-    }
-
-  return FALSE;
-}
-
 /* Both server and repository support the capability. */
 static const char *capability_yes = "yes";
 /* Either server or repository does not support the capability. */
@@ -135,26 +132,39 @@ static const char *capability_no = "no";
 static const char *capability_server_yes = "server-yes";
 
 
-/* Store in RAS the capabilities discovered from REQ's headers.
-   Use POOL for temporary allocation only. */
+/* Store in RAS the capabilities and other interesting tidbits of
+   information discovered from REQ's headers.  Use POOL for temporary
+   allocation only.
+
+   Also, if YOUNGEST_REV is not NULL, set *YOUNGEST_REV to the current
+   youngest revision if we can detect that from the OPTIONS exchange, or
+   to SVN_INVALID_REVNUM otherwise.  */
 static void
 parse_capabilities(ne_request *req,
                    svn_ra_neon__session_t *ras,
+                   svn_revnum_t *youngest_rev,
                    apr_pool_t *pool)
 {
-  const char *header_value;
+  const char *val;
+
+  if (youngest_rev)
+    *youngest_rev = SVN_INVALID_REVNUM;
 
   /* Start out assuming all capabilities are unsupported. */
+  apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_PARTIAL_REPLAY,
+               APR_HASH_KEY_STRING, capability_no);
   apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_DEPTH,
                APR_HASH_KEY_STRING, capability_no);
   apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_MERGEINFO,
                APR_HASH_KEY_STRING, capability_no);
   apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_LOG_REVPROPS,
                APR_HASH_KEY_STRING, capability_no);
+  apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_ATOMIC_REVPROPS,
+               APR_HASH_KEY_STRING, capability_no);
 
   /* Then find out which ones are supported. */
-  header_value = ne_get_response_header(req, "dav");
-  if (header_value)
+  val = ne_get_response_header(req, "dav");
+  if (val)
     {
       /* Multiple headers of the same name will have been merged
          together by the time we see them (either by an intermediary,
@@ -170,56 +180,123 @@ parse_capabilities(ne_request *req,
 
           Here we might see:
 
-          header_value == "1,2, version-control,checkout,working-resource, merge,baseline,activity,version-controlled-collection, http://subversion.tigris.org/xmlns/dav/svn/depth, <http://apache.org/dav/propset/fs/1>"
+          val == "1,2, version-control,checkout,working-resource, merge,baseline,activity,version-controlled-collection, http://subversion.tigris.org/xmlns/dav/svn/depth, <http://apache.org/dav/propset/fs/1>"
 
           (Deliberately not line-wrapping that, so you can see what
           we're about to parse.)
       */
 
       apr_array_header_t *vals =
-        svn_cstring_split(header_value, ",", TRUE, pool);
+        svn_cstring_split(val, ",", TRUE, pool);
 
       /* Right now we only have a few capabilities to detect, so
          just seek for them directly.  This could be written
          slightly more efficiently, but that wouldn't be worth it
          until we have many more capabilities. */
 
-      if (match_list(SVN_DAV_NS_DAV_SVN_DEPTH, vals))
+      if (svn_cstring_match_list(SVN_DAV_NS_DAV_SVN_DEPTH, vals))
         apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_DEPTH,
                      APR_HASH_KEY_STRING, capability_yes);
 
-      if (match_list(SVN_DAV_NS_DAV_SVN_MERGEINFO, vals))
+      if (svn_cstring_match_list(SVN_DAV_NS_DAV_SVN_MERGEINFO, vals))
         /* The server doesn't know what repository we're referring
            to, so it can't just say capability_yes. */
         apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_MERGEINFO,
                      APR_HASH_KEY_STRING, capability_server_yes);
 
-      if (match_list(SVN_DAV_NS_DAV_SVN_LOG_REVPROPS, vals))
+      if (svn_cstring_match_list(SVN_DAV_NS_DAV_SVN_LOG_REVPROPS, vals))
         apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_LOG_REVPROPS,
                      APR_HASH_KEY_STRING, capability_yes);
 
-      if (match_list(SVN_DAV_NS_DAV_SVN_PARTIAL_REPLAY, vals))
+      if (svn_cstring_match_list(SVN_DAV_NS_DAV_SVN_ATOMIC_REVPROPS, vals))
+        apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_ATOMIC_REVPROPS,
+                     APR_HASH_KEY_STRING, capability_yes);
+
+      if (svn_cstring_match_list(SVN_DAV_NS_DAV_SVN_PARTIAL_REPLAY, vals))
         apr_hash_set(ras->capabilities, SVN_RA_CAPABILITY_PARTIAL_REPLAY,
                      APR_HASH_KEY_STRING, capability_yes);
     }
+
+  /* Not strictly capabilities, but while we're here, we might as well... */
+  if ((val = ne_get_response_header(req, SVN_DAV_YOUNGEST_REV_HEADER)))
+    {
+      if (youngest_rev)
+        *youngest_rev = SVN_STR_TO_REV(val);
+    }
+  if ((val = ne_get_response_header(req, SVN_DAV_REPOS_UUID_HEADER)))
+    {
+      ras->uuid = apr_pstrdup(ras->pool, val);
+    }
+  if ((val = ne_get_response_header(req, SVN_DAV_ROOT_URI_HEADER)))
+    {
+      ne_uri root_uri = ras->root;
+
+      root_uri.path = (char *)val;
+      ras->repos_root = svn_ra_neon__uri_unparse(&root_uri, ras->pool);
+    }
+
+  /* HTTP v2 stuff */
+  if ((val = ne_get_response_header(req, SVN_DAV_ME_RESOURCE_HEADER)))
+    {
+#ifdef SVN_DEBUG
+      char *ignore_v2_env_var = getenv(SVN_IGNORE_V2_ENV_VAR);
+
+      if (! (ignore_v2_env_var
+             && apr_strnatcasecmp(ignore_v2_env_var, "yes") == 0))
+        ras->me_resource = apr_pstrdup(ras->pool, val);
+#else
+      ras->me_resource = apr_pstrdup(ras->pool, val);
+#endif
+    }
+  if ((val = ne_get_response_header(req, SVN_DAV_REV_ROOT_STUB_HEADER)))
+    {
+      ras->rev_root_stub = apr_pstrdup(ras->pool, val);
+    }
+  if ((val = ne_get_response_header(req, SVN_DAV_REV_STUB_HEADER)))
+    {
+      ras->rev_stub = apr_pstrdup(ras->pool, val);
+    }
+  if ((val = ne_get_response_header(req, SVN_DAV_TXN_ROOT_STUB_HEADER)))
+    {
+      ras->txn_root_stub = apr_pstrdup(ras->pool, val);
+    }
+  if ((val = ne_get_response_header(req, SVN_DAV_TXN_STUB_HEADER)))
+    {
+      ras->txn_stub = apr_pstrdup(ras->pool, val);
+    }
+  if ((val = ne_get_response_header(req, SVN_DAV_VTXN_ROOT_STUB_HEADER)))
+    {
+      ras->vtxn_root_stub = apr_pstrdup(ras->pool, val);
+    }
+  if ((val = ne_get_response_header(req, SVN_DAV_VTXN_STUB_HEADER)))
+    {
+      ras->vtxn_stub = apr_pstrdup(ras->pool, val);
+    }
 }
 
 
 svn_error_t *
 svn_ra_neon__exchange_capabilities(svn_ra_neon__session_t *ras,
+                                   const char **relocation_location,
+                                   svn_revnum_t *youngest_rev,
                                    apr_pool_t *pool)
 {
   svn_ra_neon__request_t* req;
   svn_error_t *err = SVN_NO_ERROR;
   ne_xml_parser *parser = NULL;
   options_ctx_t oc = { 0 };
-  const char *msg;
   int status_code;
 
   oc.pool = pool;
   oc.cdata = svn_stringbuf_create("", pool);
 
-  req = svn_ra_neon__request_create(ras, "OPTIONS", ras->url->data, pool);
+  if (youngest_rev)
+    *youngest_rev = SVN_INVALID_REVNUM;
+  if (relocation_location)
+    *relocation_location = NULL;
+
+  SVN_ERR(svn_ra_neon__request_create(&req, ras, "OPTIONS", ras->url->data,
+                                      pool));
 
   /* ### Use a symbolic name somewhere for this MIME type? */
   ne_add_request_header(req->ne_req, "Content-Type", "text/xml");
@@ -236,20 +313,22 @@ svn_ra_neon__exchange_capabilities(svn_ra_neon__session_t *ras,
                                            "<D:options xmlns:D=\"DAV:\">"
                                            "<D:activity-collection-set/>"
                                            "</D:options>",
-                                           200, 0, pool)))
+                                           200,
+                                           relocation_location ? 301 : 0,
+                                           pool)))
     goto cleanup;
 
-  /* Was there an XML parse error somewhere? */
-  msg = ne_xml_get_error(parser);
-  if (msg && *msg)
+  if (req->code == 301)
     {
-      err = svn_error_createf(SVN_ERR_RA_DAV_REQUEST_FAILED, NULL,
-                              _("The %s request returned invalid XML "
-                                "in the response: %s (%s)"),
-                              "OPTIONS", msg, ras->url->data);
+      *relocation_location = svn_ra_neon__request_get_location(req, pool);
       goto cleanup;
     }
 
+  /* Was there an XML parse error somewhere? */
+  err = svn_ra_neon__check_parse_error("OPTIONS", parser, ras->url->data);
+  if (err)
+    goto cleanup;
+
   /* We asked for, and therefore expect, to have found an activity
      collection in the response.  */
   if (oc.activity_coll == NULL)
@@ -262,7 +341,7 @@ svn_ra_neon__exchange_capabilities(svn_ra_neon__session_t *ras,
     }
 
   ras->act_coll = apr_pstrdup(ras->pool, oc.activity_coll->data);
-  parse_capabilities(req->ne_req, ras, pool);
+  parse_capabilities(req->ne_req, ras, youngest_rev, pool);
 
  cleanup:
   svn_ra_neon__request_destroy(req);
@@ -277,7 +356,7 @@ svn_ra_neon__get_activity_collection(const svn_string_t **activity_coll,
                                      apr_pool_t *pool)
 {
   if (! ras->act_coll)
-    SVN_ERR(svn_ra_neon__exchange_capabilities(ras, pool));
+    SVN_ERR(svn_ra_neon__exchange_capabilities(ras, NULL, NULL, pool));
   *activity_coll = svn_string_create(ras->act_coll, pool);
   return SVN_NO_ERROR;
 }
@@ -305,7 +384,9 @@ svn_ra_neon__has_capability(svn_ra_session_t *session,
 
   /* If any capability is unknown, they're all unknown, so ask. */
   if (cap_result == NULL)
-    SVN_ERR(svn_ra_neon__exchange_capabilities(ras, pool));
+    {
+      SVN_ERR(svn_ra_neon__exchange_capabilities(ras, NULL, NULL, pool));
+    }
 
 
   /* Try again, now that we've fetched the capabilities. */
@@ -313,9 +394,8 @@ svn_ra_neon__has_capability(svn_ra_session_t *session,
                             capability, APR_HASH_KEY_STRING);
 
   /* Some capabilities depend on the repository as well as the server.
-     NOTE: ../libsvn_ra_serf/serf.c:svn_ra_serf__has_capability()
-     has a very similar code block.  If you change something here,
-     check there as well. */
+     NOTE: svn_ra_serf__has_capability() has a very similar code block.  If
+     you change something here, check there as well. */
   if (cap_result == capability_server_yes)
     {
       if (strcmp(capability, SVN_RA_CAPABILITY_MERGEINFO) == 0)
@@ -326,7 +406,7 @@ svn_ra_neon__has_capability(svn_ra_session_t *session,
              above didn't even know which repository we were interested in
              -- it just told us whether the server supports mergeinfo.
              If the answer was 'no', there's no point checking the
-             particular repository; but if it was 'yes, we still must
+             particular repository; but if it was 'yes', we still must
              change it to 'no' iff the repository itself doesn't
              support mergeinfo. */
           svn_mergeinfo_catalog_t ignored;
@@ -360,9 +440,9 @@ svn_ra_neon__has_capability(svn_ra_session_t *session,
           else
             cap_result = capability_yes;
 
-          apr_hash_set(ras->capabilities,
-                       SVN_RA_CAPABILITY_MERGEINFO, APR_HASH_KEY_STRING,
-                       cap_result);
+            apr_hash_set(ras->capabilities,
+                         SVN_RA_CAPABILITY_MERGEINFO, APR_HASH_KEY_STRING,
+                         cap_result);
         }
       else
         {
