@@ -3,17 +3,22 @@
  * SVN_AUTH_CRED_SSL_CLIENT_CERT_PW
  *
  * ====================================================================
- * Copyright (c) 2000-2004, 2008 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
@@ -44,7 +49,7 @@
 #define AUTHN_PASSTYPE_KEY              "passtype"
 
 /* Baton type for the ssl client cert passphrase provider. */
-typedef struct
+typedef struct ssl_client_cert_pw_file_provider_baton_t
 {
   svn_auth_plaintext_passphrase_prompt_func_t plaintext_passphrase_prompt_func;
   void *prompt_baton;
@@ -246,8 +251,8 @@ svn_auth__ssl_client_cert_pw_file_save_creds_helper
                       apr_pool_t *cached_answer_pool;
 
                       /* Nothing cached for this realm, prompt the user. */
-                      SVN_ERR((*b->plaintext_passphrase_prompt_func)
-                               (&may_save_passphrase,
+                      SVN_ERR((*b->plaintext_passphrase_prompt_func)(
+                                &may_save_passphrase,
                                 realmstring,
                                 b->prompt_baton,
                                 pool));
@@ -400,7 +405,7 @@ svn_auth_get_ssl_client_cert_pw_file_provider2
 
 /* Baton type for client passphrase prompting.
    There is no iteration baton type. */
-typedef struct
+typedef struct ssl_client_cert_pw_prompt_provider_baton_t
 {
   svn_auth_ssl_client_cert_pw_prompt_func_t prompt_func;
   void *prompt_baton;
@@ -410,7 +415,7 @@ typedef struct
 } ssl_client_cert_pw_prompt_provider_baton_t;
 
 /* Iteration baton. */
-typedef struct
+typedef struct ssl_client_cert_pw_prompt_iter_baton_t
 {
   /* The original provider baton */
   ssl_client_cert_pw_prompt_provider_baton_t *pb;
