@@ -3,17 +3,25 @@
 #  special_tests.py:  testing special and reserved file handling
 #
 #  Subversion is a tool for revision control.
-#  See http://subversion.tigris.org for more information.
+#  See http://subversion.apache.org for more information.
 #
 # ====================================================================
-# Copyright (c) 2000-2007 CollabNet.  All rights reserved.
+#    Licensed to the Apache Software Foundation (ASF) under one
+#    or more contributor license agreements.  See the NOTICE file
+#    distributed with this work for additional information
+#    regarding copyright ownership.  The ASF licenses this file
+#    to you under the Apache License, Version 2.0 (the
+#    "License"); you may not use this file except in compliance
+#    with the License.  You may obtain a copy of the License at
 #
-# This software is licensed as described in the file COPYING, which
-# you should have received as part of this distribution.  The terms
-# are also available at http://subversion.tigris.org/license-1.html.
-# If newer versions of this license are posted there, you may use a
-# newer version instead, at your option.
+#      http://www.apache.org/licenses/LICENSE-2.0
 #
+#    Unless required by applicable law or agreed to in writing,
+#    software distributed under the License is distributed on an
+#    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+#    KIND, either express or implied.  See the License for the
+#    specific language governing permissions and limitations
+#    under the License.
 ######################################################################
 
 # General modules
@@ -25,9 +33,12 @@ import svntest
 from svntest.main import server_has_mergeinfo
 
 # (abbreviation)
-Skip = svntest.testcase.Skip
-SkipUnless = svntest.testcase.SkipUnless
-XFail = svntest.testcase.XFail
+Skip = svntest.testcase.Skip_deco
+SkipUnless = svntest.testcase.SkipUnless_deco
+XFail = svntest.testcase.XFail_deco
+Issues = svntest.testcase.Issues_deco
+Issue = svntest.testcase.Issue_deco
+Wimp = svntest.testcase.Wimp_deco
 Item = svntest.wc.StateItem
 
 
@@ -38,7 +49,7 @@ Item = svntest.wc.StateItem
 
 
 #----------------------------------------------------------------------
-
+@SkipUnless(svntest.main.is_posix_os)
 def general_symlink(sbox):
   "general symlink handling"
 
@@ -122,6 +133,7 @@ def general_symlink(sbox):
                                         expected_status, None, wc_dir)
 
 
+@SkipUnless(svntest.main.is_posix_os)
 def replace_file_with_symlink(sbox):
   "replace a normal file with a special file"
 
@@ -145,7 +157,7 @@ def replace_file_with_symlink(sbox):
                                                                'log msg',
                                                                wc_dir)
 
-  regex = 'svn: Commit failed'
+  regex = 'svn: E145001: Commit failed'
   for line in stderr_lines:
     if re.match(regex, line):
       break
@@ -153,6 +165,7 @@ def replace_file_with_symlink(sbox):
     raise svntest.Failure
 
 
+@SkipUnless(svntest.main.is_posix_os)
 def import_export_symlink(sbox):
   "import and export a symlink"
 
@@ -209,7 +222,8 @@ def import_export_symlink(sbox):
 
 #----------------------------------------------------------------------
 # Regression test for issue 1986
-
+@Issue(1986)
+@SkipUnless(svntest.main.is_posix_os)
 def copy_tree_with_symlink(sbox):
   "'svn cp dir1 dir2' which contains a symlink"
 
@@ -247,12 +261,16 @@ def copy_tree_with_symlink(sbox):
     'A/D/H2/chi' : Item(status='  ', copied='+', wc_rev='-'),
     'A/D/H2/omega' : Item(status='  ', copied='+', wc_rev='-'),
     'A/D/H2/psi' : Item(status='  ', copied='+', wc_rev='-'),
-    'A/D/H2/linktarget' : Item(status='  ', copied='+', wc_rev='-'),
-    'A/D/H2/newfile' : Item(status='  ', copied='+', wc_rev='-'),
+    # linktarget and newfile are from r2, while h2 is from r1.
+    'A/D/H2/linktarget' : Item(status='A ', copied='+', wc_rev='-',
+                               entry_status='  '),
+    'A/D/H2/newfile' : Item(status='A ', copied='+', wc_rev='-',
+                            entry_status='  '),
     })
   svntest.actions.run_and_verify_status(wc_dir, expected_status)
 
 
+@SkipUnless(svntest.main.is_posix_os)
 def replace_symlink_with_file(sbox):
   "replace a special file with a non-special file"
 
@@ -297,7 +315,7 @@ def replace_symlink_with_file(sbox):
                                                                'log msg',
                                                                wc_dir)
 
-  regex = 'svn: Commit failed'
+  regex = 'svn: E145001: Commit failed'
   for line in stderr_lines:
     if re.match(regex, line):
       break
@@ -305,6 +323,7 @@ def replace_symlink_with_file(sbox):
     raise svntest.Failure
 
 
+@SkipUnless(svntest.main.is_posix_os)
 def remove_symlink(sbox):
   "remove a symlink"
 
@@ -348,6 +367,9 @@ def remove_symlink(sbox):
   svntest.actions.run_and_verify_commit(wc_dir, expected_output,
                                         expected_status, None, wc_dir)
 
+@SkipUnless(svntest.main.is_posix_os)
+@SkipUnless(server_has_mergeinfo)
+@Issue(2530)
 def merge_symlink_into_file(sbox):
   "merge symlink into file"
 
@@ -393,7 +415,7 @@ def merge_symlink_into_file(sbox):
                        'merge', '-r', '2:4', dprime_url,
                        os.path.join(wc_dir, 'A', 'D'))
 
-  # now revert, and we'll get a strange error
+  # now revert, we once got a strange error
   svntest.main.run_svn(None, 'revert', '-R', wc_dir)
 
   # assuming we got past the revert because someone fixed that bug, lets
@@ -413,6 +435,7 @@ def merge_symlink_into_file(sbox):
 
 
 
+@SkipUnless(svntest.main.is_posix_os)
 def merge_file_into_symlink(sbox):
   "merge file into symlink"
 
@@ -465,11 +488,12 @@ def merge_file_into_symlink(sbox):
   # ok, now merge the change to the file into the symlink we created, this
   # gives us a weird error
   svntest.main.run_svn(None,
-                       'merge', '-r', '4:5', d_url,
+                       'merge', '-r', '4:5', '--allow-mixed-revisions', d_url,
                        os.path.join(wc_dir, 'A', 'Dprime'))
 
 # Issue 2701: Tests to see repository with symlinks can be checked out on all
 # platforms.
+@Issue(2701)
 def checkout_repo_with_symlinks(sbox):
   "checkout a repository containing symlinks"
 
@@ -497,6 +521,8 @@ def checkout_repo_with_symlinks(sbox):
                                           expected_wc)
 
 # Issue 2716: 'svn diff' against a symlink to a directory within the wc
+@Issue(2716)
+@SkipUnless(svntest.main.is_posix_os)
 def diff_symlink_to_dir(sbox):
   "diff a symlink to a directory"
 
@@ -516,7 +542,7 @@ def diff_symlink_to_dir(sbox):
     "Index: link\n",
     "===================================================================\n",
     "--- link\t(revision 0)\n",
-    "+++ link\t(revision 0)\n",
+    "+++ link\t(working copy)\n",
     "@@ -0,0 +1 @@\n",
     "+link " + d_path + "\n",
     "\ No newline at end of file\n",
@@ -524,8 +550,10 @@ def diff_symlink_to_dir(sbox):
     "Property changes on: link\n",
     "___________________________________________________________________\n",
     "Added: svn:special\n",
-    "   + *\n",
-    "\n" ]
+    "## -0,0 +1 ##\n",
+    "+*\n",
+    "\\ No newline at end of property\n"
+  ]
   svntest.actions.run_and_verify_svn(None, expected_output, [], 'diff',
                                      '.')
   # We should get the same output if we the diff the symlink itself.
@@ -534,6 +562,7 @@ def diff_symlink_to_dir(sbox):
 
 # Issue 2692 (part of): Check that the client can check out a repository
 # that contains an unknown special file type.
+@Issue(2692)
 def checkout_repo_with_unknown_special_type(sbox):
   "checkout repository with unknown special file type"
 
@@ -576,14 +605,20 @@ def replace_symlink_with_dir(sbox):
   # I'd expect a failed commit here, but replacing a file locally with a
   # directory seems to make svn think the file is unchanged.
   os.chdir(was_cwd)
-  exit_code, stdout_lines, stderr_lines = svntest.main.run_svn(1, 'ci', '-m',
-                                                               'log msg',
-                                                               wc_dir)
-  if not (stdout_lines == [] or stderr_lines == []):
-    raise svntest.Failure
+  expected_output = svntest.wc.State(wc_dir, {
+  })
+
+  if svntest.main.is_posix_os():
+    error_re_string = '.*E145001: Entry.*has unexpectedly changed special.*'
+  else:
+    error_re_string = None
+  svntest.actions.run_and_verify_commit(wc_dir, expected_output,
+                                        None, error_re_string, wc_dir)
 
 # test for issue #1808: svn up deletes local symlink that obstructs
 # versioned file
+@Issue(1808)
+@SkipUnless(svntest.main.is_posix_os)
 def update_obstructing_symlink(sbox):
   "symlink obstructs incoming delete"
 
@@ -612,23 +647,55 @@ def update_obstructing_symlink(sbox):
 def warn_on_reserved_name(sbox):
   "warn when attempt operation on a reserved name"
   sbox.build()
-  wc_dir = sbox.wc_dir
-  if os.path.exists(os.path.join(wc_dir, ".svn")):
-    reserved_path = os.path.join(wc_dir, ".svn")
-  elif os.path.exists(os.path.join(wc_dir, "_svn")):
-    reserved_path = os.path.join(wc_dir, "_svn")
-  else:
-    # We don't know how to test this, but have no reason to believe
-    # it would fail.  (TODO: any way to return 'Skip', though?)
-    return
+  reserved_path = os.path.join(sbox.wc_dir, svntest.main.get_admin_name())
   svntest.actions.run_and_verify_svn(
     "Locking a file with a reserved name failed to result in an error",
     None,
-    ".*Skipping argument: '.+' ends in a reserved name.*",
+    ".*Skipping argument: E200025: '.+' ends in a reserved name.*",
     'lock', reserved_path)
 
 
+def propvalue_normalized(sbox):
+  "'ps svn:special' should normalize to '*'"
+
+  sbox.build()
+  wc_dir = sbox.wc_dir
+
+  # Add a "symlink"
+  iota2_path = sbox.ospath('iota2')
+  svntest.main.file_write(iota2_path, "symlink destination")
+  svntest.main.run_svn(None, 'add', iota2_path)
+  svntest.main.run_svn(None, 'propset', 'svn:special', 'yes', iota2_path)
+  if svntest.main.is_posix_os():
+    os.remove(iota2_path)
+    os.symlink("symlink destination", iota2_path)
+
+  # Property value should be SVN_PROP_BOOLEAN_TRUE
+  expected_propval = ['*']
+  svntest.actions.run_and_verify_svn(None, expected_propval, [],
+                                     'propget', '--strict', 'svn:special',
+                                     iota2_path)
+
+  # Commit and check again.
+  expected_output = svntest.wc.State(wc_dir, {
+    'iota2' : Item(verb='Adding'),
+    })
+  expected_status = svntest.actions.get_virginal_state(wc_dir, 1)
+  expected_status.add({
+    'iota2' : Item(status='  ', wc_rev=2),
+    })
+  svntest.actions.run_and_verify_commit(wc_dir, expected_output,
+                                        expected_status, None,
+                                        wc_dir)
+
+  svntest.main.run_svn(None, 'update', wc_dir)
+  svntest.actions.run_and_verify_svn(None, expected_propval, [],
+                                     'propget', '--strict', 'svn:special',
+                                     iota2_path)
+
+
 # on users@: http://mid.gmane.org/1292856447.8650.24.camel@nimble.325Bayport
+@SkipUnless(svntest.main.is_posix_os)
 def unrelated_changed_special_status(sbox):
   "commit foo while bar changed special status"
 
@@ -646,30 +713,258 @@ def unrelated_changed_special_status(sbox):
                                      '-m', 'psi changed special status')
 
 
+@Issue(3972)
+@SkipUnless(svntest.main.is_posix_os)
+def symlink_destination_change(sbox):
+  "revert a symlink destination change"
+
+  sbox.build()
+  wc_dir = sbox.wc_dir
+
+  # Create a new symlink and commit it.
+  newfile_path = os.path.join(wc_dir, 'newfile')
+  os.symlink('linktarget', newfile_path)
+  svntest.main.run_svn(None, 'add', newfile_path)
+
+  expected_output = svntest.wc.State(wc_dir, {
+    'newfile' : Item(verb='Adding'),
+    })
+
+  expected_status = svntest.actions.get_virginal_state(wc_dir, 1)
+  expected_status.add({
+    'newfile' : Item(status='  ', wc_rev=2),
+    })
+
+  svntest.actions.run_and_verify_commit(wc_dir, expected_output,
+                                        expected_status, None, wc_dir)
+
+  # Modify the symlink to point somewhere else
+  os.remove(newfile_path)
+  os.symlink('linktarget2', newfile_path)
+
+  expected_status.tweak('newfile', status='M ')
+  svntest.actions.run_and_verify_status(wc_dir, expected_status)
+
+  # Revert should restore the symlink to point to the original destination
+  svntest.main.run_svn(None, 'revert', '-R', wc_dir)
+  expected_status.tweak('newfile', status='  ')
+  svntest.actions.run_and_verify_status(wc_dir, expected_status)
+
+  # Issue 3972, repeat revert produces no output
+  svntest.actions.run_and_verify_svn(None, [], [], 'revert', '-R', wc_dir)
+  svntest.actions.run_and_verify_status(wc_dir, expected_status)
+
+  # Now replace the symlink with a normal file and try to commit, we
+
+#----------------------------------------------------------------------
+# This used to lose the special status in the target working copy
+# (disk and metadata).
+@Issue(3884)
+@SkipUnless(svntest.main.is_posix_os)
+@XFail()
+def merge_foreign_symlink(sbox):
+  "merge symlink-add from foreign repos"
+
+  sbox.build()
+  wc_dir = sbox.wc_dir
+
+  # Make a copy of this repository and associated working copy.  Both
+  # should have nothing but a Greek tree in them, and the two
+  # repository UUIDs should differ.
+  sbox2 = sbox.clone_dependent(True)
+  sbox2.build()
+  wc_dir2 = sbox2.wc_dir
+
+  # convenience variables
+  zeta_path = sbox.ospath('A/zeta')
+  zeta2_path = sbox2.ospath('A/zeta')
+
+  # sbox2 r2: create zeta2 in sbox2
+  os.symlink('target', zeta2_path)
+  sbox2.simple_add('A/zeta')
+  sbox2.simple_commit('A/zeta')
+
+
+  # sbox1: merge that
+  svntest.main.run_svn(None, 'merge', '-c', '2', sbox2.repo_url,
+                       sbox.ospath(''))
+
+  # Verify special status.
+  expected_disk = svntest.main.greek_state.copy()
+  expected_disk.add({
+    'A/zeta': Item(props={ 'svn:special': '*' })
+  })
+  svntest.actions.verify_disk(sbox.ospath(''), expected_disk, True)
+
+  # TODO: verify status:
+  #   expected_status = svntest.actions.get_virginal_state(wc_dir, 2)
+  #   expected_status.add({
+  #     'A/zeta' : Item(status='A ', wc_rev='-', props={'svn:special': '*'}),
+  #     })
+
+#----------------------------------------------------------------------
+# See also symlink_to_wc_svnversion().
+@Issue(2557,3987)
+@SkipUnless(svntest.main.is_posix_os)
+def symlink_to_wc_basic(sbox):
+  "operate on symlink to wc"
+
+  sbox.build(read_only = True)
+  wc_dir = sbox.wc_dir
+
+  # Create a symlink
+  symlink_path = sbox.add_wc_path('2')
+  assert not os.path.islink(symlink_path)
+  os.symlink(os.path.basename(wc_dir), symlink_path) ### implementation detail
+  symlink_basename = os.path.basename(symlink_path)
+
+  # Some basic tests
+  wc_uuid = svntest.actions.get_wc_uuid(wc_dir)
+  expected_info = [{
+      'Path' : re.escape(os.path.join(symlink_path)),
+      'Working Copy Root Path' : re.escape(os.path.abspath(symlink_path)),
+      'Repository Root' : sbox.repo_url,
+      'Repository UUID' : wc_uuid,
+      'Revision' : '1',
+      'Node Kind' : 'directory',
+      'Schedule' : 'normal',
+  }, {
+      'Name' : 'iota',
+      'Path' : re.escape(os.path.join(symlink_path, 'iota')),
+      'Working Copy Root Path' : re.escape(os.path.abspath(symlink_path)),
+      'Repository Root' : sbox.repo_url,
+      'Repository UUID' : wc_uuid,
+      'Revision' : '1',
+      'Node Kind' : 'file',
+      'Schedule' : 'normal',
+  }]
+  svntest.actions.run_and_verify_info(expected_info,
+                                      symlink_path, symlink_path + '/iota')
+
+#----------------------------------------------------------------------
+# Similar to #2557/#3987; see symlink_to_wc_basic().
+@Issue(2557,3987)
+@SkipUnless(svntest.main.is_posix_os)
+def symlink_to_wc_svnversion(sbox):
+  "svnversion on symlink to wc"
+
+  sbox.build(read_only = True)
+  wc_dir = sbox.wc_dir
+
+  # Create a symlink
+  symlink_path = sbox.add_wc_path('2')
+  assert not os.path.islink(symlink_path)
+  os.symlink(os.path.basename(wc_dir), symlink_path) ### implementation detail
+  symlink_basename = os.path.basename(symlink_path)
+
+  # Some basic tests
+  svntest.actions.run_and_verify_svnversion("Unmodified symlink to wc",
+                                            symlink_path, sbox.repo_url,
+                                            [ "1\n" ], [])
+
+# Regression in 1.7.0: Update fails to change a symlink
+@SkipUnless(svntest.main.is_posix_os)
+def update_symlink(sbox):
+  "update a symlink"
+
+  svntest.actions.do_sleep_for_timestamps()
+
+  sbox.build()
+  wc_dir = sbox.wc_dir
+  mu_path = sbox.ospath('A/mu')
+  iota_path = sbox.ospath('iota')
+  symlink_path = sbox.ospath('symlink')
+
+  # create a symlink to /A/mu
+  os.symlink("A/mu", symlink_path)
+  sbox.simple_add('symlink')
+  sbox.simple_commit()
+
+  # change the symlink to /iota
+  os.remove(symlink_path)
+  os.symlink("iota", symlink_path)
+  sbox.simple_commit()
+
+  # update back to r2
+  svntest.main.run_svn(False, 'update', '-r', '2', wc_dir)
+
+  # now update to head; 1.7.0 throws an assertion here
+  expected_output = svntest.wc.State(wc_dir, {
+    'symlink'          : Item(status='U '),
+  })
+  expected_disk = svntest.main.greek_state.copy()
+  expected_disk.add({'symlink': Item(contents="This is the file 'iota'.\n",
+                                     props={'svn:special' : '*'})})
+  expected_status = svntest.actions.get_virginal_state(wc_dir, 3)
+  expected_status.add({
+    'symlink'           : Item(status='  ', wc_rev='3'),
+  })
+  svntest.actions.run_and_verify_update(wc_dir,
+                                        expected_output,
+                                        expected_disk,
+                                        expected_status,
+                                        None, None, None,
+                                        None, None, 1)
+
+@Issue(4102)
+@SkipUnless(svntest.main.is_posix_os)
+def externals_as_symlink_targets(sbox):
+  "externals as symlink targets"
+  sbox.build()
+  wc = sbox.ospath
+
+  # Control: symlink to normal dir and file.
+  os.symlink('E', wc('sym_E'))
+  os.symlink('mu', wc('sym_mu'))
+
+  # Test case: symlink to external dir and file.
+  sbox.simple_propset("svn:externals",
+                      '^/A/B/E ext_E\n'
+                      '^/A/mu ext_mu',
+                      '')
+  sbox.simple_update()
+  os.symlink('ext_E', wc('sym_ext_E'))
+  os.symlink('ext_mu', wc('sym_ext_mu'))
+
+  # Adding symlinks to normal items and to a file external is OK.
+  sbox.simple_add('sym_E', 'sym_mu', 'sym_ext_mu')
+
+  ### Adding a symlink to an external dir failed with
+  ###   svn: E200009: Could not add all targets because some targets are
+  ###   already versioned
+  sbox.simple_add('sym_ext_E')
+
+  sbox.simple_commit()
+    
+
 ########################################################################
 # Run the tests
 
 
 # list all tests here, starting with None:
 test_list = [ None,
-              SkipUnless(general_symlink, svntest.main.is_posix_os),
-              SkipUnless(replace_file_with_symlink, svntest.main.is_posix_os),
-              SkipUnless(import_export_symlink, svntest.main.is_posix_os),
-              SkipUnless(copy_tree_with_symlink, svntest.main.is_posix_os),
-              SkipUnless(replace_symlink_with_file, svntest.main.is_posix_os),
-              SkipUnless(remove_symlink, svntest.main.is_posix_os),
-              SkipUnless(SkipUnless(merge_symlink_into_file,
-                                    svntest.main.is_posix_os),
-                         server_has_mergeinfo),
-              SkipUnless(merge_file_into_symlink, svntest.main.is_posix_os),
+              general_symlink,
+              replace_file_with_symlink,
+              import_export_symlink,
+              copy_tree_with_symlink,
+              replace_symlink_with_file,
+              remove_symlink,
+              merge_symlink_into_file,
+              merge_file_into_symlink,
               checkout_repo_with_symlinks,
-              XFail(SkipUnless(diff_symlink_to_dir, svntest.main.is_posix_os)),
+              diff_symlink_to_dir,
               checkout_repo_with_unknown_special_type,
               replace_symlink_with_dir,
-              SkipUnless(update_obstructing_symlink, svntest.main.is_posix_os),
+              update_obstructing_symlink,
               warn_on_reserved_name,
-              SkipUnless(unrelated_changed_special_status,
-                         svntest.main.is_posix_os),
+              propvalue_normalized,
+              unrelated_changed_special_status,
+              symlink_destination_change,
+              merge_foreign_symlink,
+              symlink_to_wc_basic,
+              symlink_to_wc_svnversion,
+              update_symlink,
+              externals_as_symlink_targets,
              ]
 
 if __name__ == '__main__':
